/*******************************************************************************
 * Copyright (c) 2011 Michael Ruflin, André Locher, Claudia von Bastian.
 * 
 * This file is part of Tatool.
 * 
 * Tatool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or 
 * (at your option) any later version.
 * 
 * Tatool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Tatool. If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/
package ch.tatool.examples.executable;

import java.awt.event.KeyEvent;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import javax.swing.SwingUtilities;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import au.com.bytecode.opencsv.CSVReader;

import ch.tatool.core.data.DataUtils;
import ch.tatool.core.data.IntegerProperty;
import ch.tatool.core.data.Misc;
import ch.tatool.core.data.Points;
import ch.tatool.core.data.Question;
import ch.tatool.core.data.Result;
import ch.tatool.core.data.StringProperty;
import ch.tatool.core.data.Timing;
import ch.tatool.core.display.swing.ExecutionDisplayUtils;
import ch.tatool.core.display.swing.SwingExecutionDisplay;
import ch.tatool.core.display.swing.action.ActionPanel;
import ch.tatool.core.display.swing.action.ActionPanelListener;
import ch.tatool.core.display.swing.action.KeyActionPanel;
import ch.tatool.core.display.swing.container.ContainerUtils;
import ch.tatool.core.display.swing.container.RegionsContainer;
import ch.tatool.core.display.swing.container.RegionsContainer.Region;
import ch.tatool.core.display.swing.panel.CenteredTextPanel;
import ch.tatool.core.executable.BlockingAWTExecutable;
import ch.tatool.data.DescriptivePropertyHolder;
import ch.tatool.data.Property;
import ch.tatool.data.Trial;
import ch.tatool.exec.ExecutionContext;
import ch.tatool.exec.ExecutionOutcome;
import ch.tatool.exec.ExecutionPhase;
import ch.tatool.exec.ExecutionPhaseListener;

/**
 * Runs a typical memory updating paradigm: First, 1 to 3 digits in different
 * colours are shown simultaneously and have to be memorised. Arithmetic
 * operations shown next have to be applied to the digits, if (and only if) they
 * are of identical colour. After each arithmetic operation, the correct
 * result has to be entered (a number between 0 and 9), and all following
 * arithmetic operations in this colour have to be applied to this result.
 * 
 * @author Claudia von Bastian
 */

public class MemoryUpdating extends BlockingAWTExecutable implements
		ActionPanelListener, DescriptivePropertyHolder, ExecutionPhaseListener {

	Logger logger = LoggerFactory.getLogger(MemoryUpdating.class);

	private CenteredTextPanel questionPanel;
	private KeyActionPanel actionPanel;
	private int correctResponse;
	private int givenResponse;
	private long startTime;
	private long endTime;
	private Timer timer;
	private TimerTask displayProcessingTask;
	private int displayItemDuration = 1000;
	private int displayCrossDuration = 2000;

	private static String STIMULI_PATH = "/ch/tatool/examples/executable/stimuli/";
	private String stimuliFile = "memory_updating_stimuli.csv";

	private List<String[]> stimuliList;
	private int itemno = 0;
	private int stimulusType;
	private int block;
	private int trialno;
	private int load;
	private String type = "";
	private String question = "";

	private StringProperty typeProperty = new StringProperty("type");
	private IntegerProperty loadProperty = new IntegerProperty("load");
	private IntegerProperty blockProperty = new IntegerProperty("block");
	private IntegerProperty trialnoProperty = new IntegerProperty("trial");

	private RegionsContainer regionsContainer;

	public MemoryUpdating() {
		questionPanel = new CenteredTextPanel();
		actionPanel = new KeyActionPanel();
		actionPanel.addActionPanelListener(this);
		timer = new Timer();
	}

	protected void startExecutionAWT() {
		
		// initialise environment for executable
		ExecutionContext context = getExecutionContext();
		SwingExecutionDisplay display = ExecutionDisplayUtils.getDisplay(context);
		ContainerUtils.showRegionsContainer(display);
		regionsContainer = ContainerUtils.getRegionsContainer();

		String stimulus = initStimulus();

		if (stimulusType == 0) {
			questionPanel.setTextSize(150);
			questionPanel.setText(stimulus);
			actionPanel.addKey(KeyEvent.VK_0, "0", 0);
			actionPanel.addKey(KeyEvent.VK_1, "1", 1);
			actionPanel.addKey(KeyEvent.VK_2, "2", 2);
			actionPanel.addKey(KeyEvent.VK_3, "3", 3);
			actionPanel.addKey(KeyEvent.VK_4, "4", 4);
			actionPanel.addKey(KeyEvent.VK_5, "5", 5);
			actionPanel.addKey(KeyEvent.VK_6, "6", 6);
			actionPanel.addKey(KeyEvent.VK_7, "7", 7);
			actionPanel.addKey(KeyEvent.VK_8, "8", 8);
			actionPanel.addKey(KeyEvent.VK_9, "9", 9);

			regionsContainer.setRegionContent(Region.CENTER, questionPanel);
			regionsContainer.setRegionContent(Region.SOUTH, actionPanel);

			Timing.getStartTimeProperty().setValue(this, new Date());
			startTime = System.nanoTime();

			regionsContainer.setRegionContentVisibility(Region.CENTER, true);
			regionsContainer.setRegionContentVisibility(Region.SOUTH, false);

			actionPanel.enableActionPanel();
			
		} else if (stimulusType == 1) {

			questionPanel.setTextSize(150);
			questionPanel.setText(stimulus);
			regionsContainer.setRegionContent(Region.CENTER, questionPanel);
			regionsContainer.setRegionContentVisibility(Region.CENTER, true);

			// timer
			displayProcessingTask = new TimerTask() {
				public void run() {
					SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							givenResponse = 0;
							endTask();
						}
					});
				}
			};

			timer.schedule(displayProcessingTask, displayItemDuration);
			
		} else {

			questionPanel.setTextSize(100);
			questionPanel.setText(stimulus);
			regionsContainer.setRegionContent(Region.CENTER, questionPanel);
			regionsContainer.setRegionContentVisibility(Region.CENTER, true);

			// timer
			displayProcessingTask = new TimerTask() {
				public void run() {
					SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							givenResponse = 0;
							endTask();
						}
					});
				}
			};

			timer.schedule(displayProcessingTask, displayCrossDuration);
		}
	}

	/**
	 * Generates the stimuli.
	 */
	private String initStimulus() {

		if (stimuliList.size() == 0) {
			readInputData();
		}

		String stimulus = "";
		String stimno = "";
		String color1 = "";
		String color2 = "";
		String color3 = "";
		String digit1 = "";
		String digit2 = "";
		String digit3 = "";

		String[] currentItem = stimuliList.get(itemno);

		// get relevant properties
		block = Integer.valueOf(currentItem[0]);
		load = Integer.valueOf(currentItem[1]);
		displayItemDuration = 1000 * load;
		trialno = Integer.valueOf(currentItem[2]);
		stimno = currentItem[3];
		type = currentItem[4];
		color1 = currentItem[5];
		color2 = currentItem[8];
		color3 = currentItem[10];
		digit1 = currentItem[6];
		digit2 = currentItem[9];
		digit3 = currentItem[11];

		// set correct response and property values
		try {
			correctResponse = Integer.valueOf(currentItem[7]);
		} catch (NumberFormatException e) {
			correctResponse = 0;
		}

		loadProperty.setValue(this, load);
		typeProperty.setValue(this, type);
		blockProperty.setValue(this, block);
		trialnoProperty.setValue(this, trialno);

		// define stimulus type; 0 = arithmetic operation, 1 = memoranda
		if (stimno.equals("0")) {
			stimulusType = 1;
		} else if (stimno.equals("-1")) {
			stimulusType = -1;
		} else {
			stimulusType = 0;
		}

		// arrange stimuli according to load
		if (stimulusType == 1) {
			switch (load) {
			case 1:
				stimulus = "<html>" + getHtmlString(digit1, color1) + "</html>";
				question = digit1;
				break;
			case 2:
				stimulus = "<html>" + getHtmlString(digit1, color1)
						+ "&nbsp;&nbsp;&nbsp;" + getHtmlString(digit2, color2)
						+ "</html>";
				question = digit1 + ";" + digit2;
				break;
			case 3:
				stimulus = "<html>" + getHtmlString(digit1, color1)
						+ "&nbsp;&nbsp;&nbsp;" + getHtmlString(digit2, color2)
						+ "&nbsp;&nbsp;&nbsp;" + getHtmlString(digit3, color3)
						+ "</html>";
				question = digit1 + ";" + digit2 + ";" + digit3;
				break;
			default:
				break;
			}
		} else if (stimulusType == -1) {
			stimulus = "+";
			question = "preparation";
		} else {
			stimulus = "<html>" + getHtmlString(digit1, color1) + "</html>";
			question = digit1;
		}

		itemno = itemno + 1;

		return stimulus;
	}
	
	/**
	 * Reads the stimuli file CSV and stores the stimuli in an ArrayList
	 */
	private void readInputData() {
		stimuliList = new ArrayList<String[]>();
		CSVReader reader = null;
		try {
			reader = new CSVReader(new InputStreamReader(this.getClass()
					.getResourceAsStream(STIMULI_PATH + stimuliFile),
					"ISO-8859-1"), ';');
		} catch (UnsupportedEncodingException e1) {
			logger.error(e1.getMessage(), e1);
		}

		try {
			stimuliList = reader.readAll();
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		} finally {
			try {
				reader.close();
			} catch (IOException e) {
				logger.error(e.getMessage(), e);
			}
		}
	}

	/**
	 * Provides an HTML string of a given digit in a given color.
	 * 
	 * @param digit the digit to display as an HTML string
	 * @param color the color to be used for the HTML string
	 * @return
	 */
	private String getHtmlString(String digit, String color) {
		String htmlString = "";

		if (color.equals("lightblue")) {
			htmlString = "<font color='#6ecdf5'>" + digit + "</font>";
		} else if (color.equals("orange")) {
			htmlString = "<font color='#f59100'>" + digit + "</font>";
		} else {
			htmlString = "<font color='#5f5aaa'>" + digit + "</font>";
		}

		return htmlString;
	}

	public void actionTriggered(ActionPanel source, Object actionValue) {
		actionPanel.disableActionPanel();

		endTime = System.nanoTime();
		Timing.getEndTimeProperty().setValue(this, new Date());

		givenResponse = (Integer) actionValue;

		endTask();
	}

	private void endTask() {
		regionsContainer.setRegionContentVisibility(Region.CENTER, false);
		regionsContainer.setRegionContentVisibility(Region.SOUTH, false);
		regionsContainer.removeRegionContent(Region.CENTER);
		regionsContainer.removeRegionContent(Region.SOUTH);

		Question.getResponseProperty().setValue(this, givenResponse);
		Question.setQuestionAnswer(this, question, correctResponse);
		boolean success = correctResponse == givenResponse;
		Points.setZeroOneMinMaxPoints(this);
		Points.setZeroOnePoints(this, success);
		Result.getResultProperty().setValue(this, success);
		Misc.getOutcomeProperty().setValue(this, ExecutionOutcome.FINISHED);

		// set duration time property
		long duration = 0;
		if (endTime > 0) {
			duration = endTime - startTime;
		}
		long ms = (long) duration / 1000000;
		Timing.getDurationTimeProperty().setValue(this, ms);

		// create new trial
		Trial currentTrial = getExecutionContext().getExecutionData().addTrial();
		currentTrial.setParentId(getId());

		// add all executable properties to the current trial
		DataUtils.storeProperties(currentTrial, this);

		// finish the execution and make sure nobody else already did so
		if (getFinishExecutionLock()) {
			finishExecution();
		}
	}

	/**
	 * Is called whenever we copy the properties from our executable to a trial
	 * object for persistence with the help of the DataUtils class.
	 */
	public Property<?>[] getPropertyObjects() {
		return new Property[] { Points.getMinPointsProperty(),
				Points.getPointsProperty(), Points.getMaxPointsProperty(),
				Question.getQuestionProperty(), Question.getAnswerProperty(),
				Question.getResponseProperty(), Result.getResultProperty(),
				Timing.getStartTimeProperty(), Timing.getEndTimeProperty(),
				Timing.getDurationTimeProperty(), Misc.getOutcomeProperty(),
				typeProperty, loadProperty, blockProperty, trialnoProperty };
	}


	/**
	 * Is called whenever the Tatool execution phase changes. We use the
	 * SESSION_START phase to read our stimuli list.
	 */
	public void processExecutionPhase(ExecutionContext context) {
		if (context.getPhase().equals(ExecutionPhase.SESSION_START)) {
			readInputData();
		}
	}
	
	protected void cancelExecutionAWT() {
		timer.cancel();
    }

	/**
	 * Allows to set the stimuli file as a property in the module xml file.
	 * 
	 * @param stimuliFile
	 */
	public void setStimuliFile(String stimuliFile) {
		this.stimuliFile = stimuliFile;
	}

	public String getStimuliFile() {
		return stimuliFile;
	}
}
