/*******************************************************************************
 * Copyright (c) 2011 Michael Ruflin, André Locher, Claudia von Bastian.
 * 
 * This file is part of Tatool.
 * 
 * Tatool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or 
 * (at your option) any later version.
 * 
 * Tatool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Tatool. If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/
package ch.tatool.examples.executable;

import java.awt.event.KeyEvent;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import javax.swing.SwingUtilities;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import au.com.bytecode.opencsv.CSVReader;

import ch.tatool.core.data.DataUtils;
import ch.tatool.core.data.IntegerProperty;
import ch.tatool.core.data.Misc;
import ch.tatool.core.data.Points;
import ch.tatool.core.data.Question;
import ch.tatool.core.data.Result;
import ch.tatool.core.data.StringProperty;
import ch.tatool.core.data.Timing;
import ch.tatool.core.display.swing.ExecutionDisplayUtils;
import ch.tatool.core.display.swing.SwingExecutionDisplay;
import ch.tatool.core.display.swing.action.ActionPanel;
import ch.tatool.core.display.swing.action.ActionPanelListener;
import ch.tatool.core.display.swing.action.KeyActionPanel;
import ch.tatool.core.display.swing.container.ContainerUtils;
import ch.tatool.core.display.swing.container.RegionsContainer;
import ch.tatool.core.display.swing.container.RegionsContainer.Region;
import ch.tatool.core.display.swing.panel.CenteredTextPanel;
import ch.tatool.core.executable.BlockingAWTExecutable;
import ch.tatool.data.DescriptivePropertyHolder;
import ch.tatool.data.Property;
import ch.tatool.data.Trial;
import ch.tatool.exec.ExecutionContext;
import ch.tatool.exec.ExecutionOutcome;
import ch.tatool.exec.ExecutionPhase;
import ch.tatool.exec.ExecutionPhaseListener;

/**
 * Runs a simple n-back test displaying letters one at a time and requiring the user to 
 * respond to every letter whether the same letter appeared before n letters.
 * 
 * @author André Locher
 */
public class NBack extends BlockingAWTExecutable implements
		ActionPanelListener, ExecutionPhaseListener, DescriptivePropertyHolder {

	Logger logger = LoggerFactory.getLogger(NBack.class);

	private StringProperty typeProperty = new StringProperty("type");
	private IntegerProperty loadProperty = new IntegerProperty("load");
	private IntegerProperty blockProperty = new IntegerProperty("block");

	/** Timing and task display duration */
	private Timer timer;
	private TimerTask taskStimuli;
	private TimerTask taskEnd;
	private int displayDuration = 500;
	private int taskDuration = 3000;

	/** Question panel. */
	private CenteredTextPanel questionPanel;
	private KeyActionPanel actionPanel;

	/** Stimuli */
	private List<String[]> stimuliList;
	private int itemno = 0;
	private int correctResponse;
	private int givenResponse;
	private String question = "";
	private String type ="";
	private int load;
	private int block;

	private long startTime;
	private long endTime;

	public static final String STIMULI_PATH = "/ch/tatool/examples/executable/stimuli/";
	public String stimuliFile = "2back_test.csv";

	private RegionsContainer regionsContainer;

	/** Default Constructor. */
	public NBack() {
		questionPanel = new CenteredTextPanel();
		actionPanel = new KeyActionPanel();
		actionPanel.addActionPanelListener(this);
		timer = new Timer();
	}

	public void startExecutionAWT() {
		
		// initialise environment for executable
		ExecutionContext context = getExecutionContext();
		SwingExecutionDisplay display = ExecutionDisplayUtils.getDisplay(context);
		ContainerUtils.showRegionsContainer(display);
		regionsContainer = ContainerUtils.getRegionsContainer();

		String stimulus = initStimulus();

		// setup the display
		questionPanel.setTextSize(100);
		questionPanel.setText(stimulus);

		actionPanel.addKey(KeyEvent.VK_A, "yes", 1); // yes
		actionPanel.addKey(KeyEvent.VK_L, "no", 0); // no

		regionsContainer.setRegionContent(Region.CENTER, questionPanel);
		regionsContainer.setRegionContent(Region.SOUTH, actionPanel);

		Timing.getStartTimeProperty().setValue(this, new Date());
		startTime = System.nanoTime();

		regionsContainer.setRegionContentVisibility(Region.CENTER, true);
		regionsContainer.setRegionContentVisibility(Region.SOUTH, false);

		actionPanel.enableActionPanel();

		// stimuli display timer task
		taskStimuli = new TimerTask() {
			public void run() {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						clearStimuli();
					}
				});
			}
		};
		timer.schedule(taskStimuli, displayDuration);

		taskEnd = new TimerTask() {
			public void run() {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						actionPanel.disableActionPanel();
						endTask();
					}
				});
			}
		};
		timer.schedule(taskEnd, taskDuration);
	}

	/**
	 * Generates the stimuli.
	 */
	private String initStimulus() {

		if (stimuliList.size() == 0) {
			readInputData();
		}

		String[] currentItem = stimuliList.get(itemno);
		String stimulus = "";

		question = currentItem[0];
		stimulus = currentItem[0];
		correctResponse = Integer.valueOf(currentItem[1]);
		type = currentItem[2];
		load = Integer.valueOf(currentItem[3]);
		block = Integer.valueOf(currentItem[4]);

		typeProperty.setValue(this, type);
		loadProperty.setValue(this, load);
		blockProperty.setValue(this, block);

		itemno = itemno + 1;

		return stimulus;
	}

	private void readInputData() {
		stimuliList = new ArrayList<String[]>();
		CSVReader reader = null;
		try {
			reader = new CSVReader(new InputStreamReader(this.getClass()
					.getResourceAsStream(STIMULI_PATH + stimuliFile),
					"ISO-8859-1"), ';');
		} catch (UnsupportedEncodingException e1) {
			logger.error(e1.getMessage(), e1);
		}

		try {
			stimuliList = reader.readAll();
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		} finally {
			try {
				reader.close();
			} catch (IOException e) {
				logger.error(e.getMessage(), e);
			}
		}
	}

	private void clearStimuli() {
		regionsContainer.setRegionContentVisibility(Region.CENTER, false);
		regionsContainer.setRegionContentVisibility(Region.SOUTH, false);
		regionsContainer.removeRegionContent(Region.CENTER);
		regionsContainer.removeRegionContent(Region.SOUTH);
	}

	/** Called when action panel listener receives an event */
	public void actionTriggered(ActionPanel source, Object actionValue) {

		givenResponse = (Integer) actionValue;

		actionPanel.disableActionPanel();
		endTime = System.nanoTime();
		Timing.getEndTimeProperty().setValue(this, new Date());
	}

	private void endTask() {
		regionsContainer.setRegionContentVisibility(Region.CENTER, false);
		regionsContainer.setRegionContentVisibility(Region.SOUTH, false);
		regionsContainer.removeRegionContent(Region.CENTER);
		regionsContainer.removeRegionContent(Region.SOUTH);

		Question.getResponseProperty().setValue(this, givenResponse);
		Question.setQuestionAnswer(this, question, correctResponse);
		boolean success = correctResponse == givenResponse;
		Points.setZeroOneMinMaxPoints(this);
		Points.setZeroOnePoints(this, success);
		Result.getResultProperty().setValue(this, success);
		Misc.getOutcomeProperty().setValue(this, ExecutionOutcome.FINISHED);

		// set duration time property
		long duration = 0;
		if (endTime > 0) {
			duration = endTime - startTime;
		}
		long ms = (long) duration / 1000000;
		Timing.getDurationTimeProperty().setValue(this, ms);

		// create new trial
		Trial currentTrial = getExecutionContext().getExecutionData()
				.addTrial();
		currentTrial.setParentId(getId());

		// add all executable properties to the current trial
		DataUtils.storeProperties(currentTrial, this);

		// finish the execution and make sure nobody else already did so
		if (getFinishExecutionLock()) {
			finishExecution();
		}
	}

	/**
	 * Is called whenever we copy the properties from our executable to a trial
	 * object for persistence with the help of the DataUtils class.
	 */
	public Property<?>[] getPropertyObjects() {
		return new Property[] { Points.getMinPointsProperty(),
				Points.getPointsProperty(), Points.getMaxPointsProperty(),
				Question.getQuestionProperty(), Question.getAnswerProperty(),
				Question.getResponseProperty(), Result.getResultProperty(),
				Timing.getStartTimeProperty(), Timing.getEndTimeProperty(),
				Timing.getDurationTimeProperty(), Misc.getOutcomeProperty(),
				typeProperty, loadProperty, blockProperty };
	}

	/**
	 * Is called whenever the Tatool execution phase changes. We use the
	 * SESSION_START phase to read our stimuli list.
	 */
	public void processExecutionPhase(ExecutionContext context) {
		if (context.getPhase().equals(ExecutionPhase.SESSION_START)) {
			readInputData();
		}
	}
	
	protected void cancelExecutionAWT() {
		timer.cancel();
    }

	public String getStimuliFile() {
		return stimuliFile;
	}

	public void setStimuliFile(String stimuliFile) {
		this.stimuliFile = stimuliFile;
	}

}
