/*******************************************************************************
 * Copyright (c) 2011 Michael Ruflin, André Locher, Claudia von Bastian.
 * 
 * This file is part of Tatool.
 * 
 * Tatool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or 
 * (at your option) any later version.
 * 
 * Tatool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Tatool. If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/
package ch.tatool.examples.module;

import java.awt.AWTException;
import java.awt.Robot;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Timer;
import java.util.TimerTask;

import javax.swing.JFileChooser;

import org.apache.commons.io.FileUtils;

import ch.tatool.app.export.CSVTrialDataExport;
import ch.tatool.core.executable.HTMLExecutable;
import ch.tatool.core.module.creator.ModuleCreationWorker;
import ch.tatool.data.Module;
import ch.tatool.data.UserAccount;
import ch.tatool.examples.executable.MyExecutable;
import ch.tatool.exec.ExecutionContext;
import ch.tatool.exec.ExecutionPhase;
import ch.tatool.exec.ExecutionPhaseListener;
import ch.tatool.test.swing.SwingModuleRunner;

/**
 * @author jrox
 *
 */
public class DemoModuleTest extends SwingModuleRunner implements
		ExecutionPhaseListener {

	private Timer execTimer;
	private Robot robot;
	private File moduleFile;
	private UserAccount account;
	private Module module;
	private int sessionCounter;

	// custom values for testing
	private int numSessions = 1;
	private static String EXPORT_FILE_NAME = "demoModule";
	private static String MODULE_FILE_NAME = "/demoModule_test.xml";

	public static void main(String[] args) {
		new DemoModuleTest().init();
	}

	public void init() {
		// setup a timer to handle the delay of our robot input
		execTimer = new Timer();

		// initialize the application with the test context
		startApplication();

		// add test class as phase listener for the following phases
		getExecutionService().getPhaseListenerManager().addExecutionPhaseListener(this, ExecutionPhase.EXECUTE_EXECUTABLE);
		getExecutionService().getPhaseListenerManager().addExecutionPhaseListener(this, ExecutionPhase.EXECUTION_FINISH);

		// create the account and the module
		account = createAccount();
		module = createModule(account);

		// initialize the UI with the account and module and start
		getGuiController().setEnabled(true);
		getGuiController().setUserAccount(account);
		getGuiController().setModule(module);

		// start module
		runModule();
	}

	/**
	 * Robot starting a module by pressing Enter
	 */
	public void runModule() {
		sessionCounter = 0;
		try {
			robot = new Robot();
			TimerTask startModuleTask = new TimerTask() {
				public void run() {
					robot.keyPress(KeyEvent.VK_ENTER);
					robot.keyRelease(KeyEvent.VK_ENTER);
				}
			};
			execTimer.schedule(startModuleTask, 500);
		} catch (AWTException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Exporting all data as a CSV file with curent millisecond time in filename
	 */
	private void exportData() {
		File moduleDataFile = null;
		CSVTrialDataExport csvExport = new CSVTrialDataExport(getDataService());
		csvExport.setCSVParameters(';');
		File tmpFile = csvExport.exportData(module, 0);
		try {
			moduleDataFile = new File(EXPORT_FILE_NAME + "_" + System.currentTimeMillis() + ".csv");
			FileUtils.copyFile(tmpFile, moduleDataFile);
			tmpFile.delete();
		} catch (IOException ioe) {
		}
	}

	/**
	 * Stop test framework by removing test account, deleting all database data and exiting the Tatool client.
	 */
	public void cleanup() {
		execTimer.cancel();
		getUserAccountService().deleteAccount(account);
		deleteData();
		System.exit(0);
	}

	/**
	 * Creates a module from a module XML file
	 */
	protected Module createModule(UserAccount account) {
		if (MODULE_FILE_NAME != null && !MODULE_FILE_NAME.equals("")) {
			try {
				URL moduleUrl = this.getClass().getResource(MODULE_FILE_NAME);
				moduleFile = new File(moduleUrl.toURI());
			} catch (URISyntaxException e) {
				e.printStackTrace();
			}
		} else {
			JFileChooser chooser = new JFileChooser();
			int result = chooser.showOpenDialog(null);
			if (result == JFileChooser.APPROVE_OPTION) {
				moduleFile = chooser.getSelectedFile();
			}
		}
		ModuleCreationWorker worker = new ModuleCreationWorker();
		worker.createModule(getModuleService(), account, moduleFile, null,
				null, false);
		return worker.getModule();
	}

	
	/**
	 * Listener method for execution phase changes
	 */
	public void processExecutionPhase(ExecutionContext context) {
		if (context.getPhase().equals(ExecutionPhase.EXECUTE_EXECUTABLE)) {
			processExecuteExecutablePhase(context);
		} else if (context.getPhase().equals(ExecutionPhase.EXECUTION_FINISH)) {
			processExecutionFinishPhase();
		}
	}

	/**
	 * Processing of EXECUTE_EXECUTABLE phase to define the robot behaviour during the execution of an executable.
	 */
	private void processExecuteExecutablePhase(ExecutionContext context) {
		if (context.getActiveExecutable() instanceof MyExecutable) {
			TimerTask responsePhase = new TimerTask() {
				public void run() {
					// always answer by pressing Key A
					robot.keyPress(KeyEvent.VK_A);
					robot.keyRelease(KeyEvent.VK_A);
				}
			};
			execTimer.schedule(responsePhase, 200);
		} else if (context.getActiveExecutable() instanceof HTMLExecutable) {
			TimerTask responsePhase = new TimerTask() {
				public void run() {
					// always press arrow right to skip through instructions
					robot.keyPress(KeyEvent.VK_RIGHT);
					robot.keyRelease(KeyEvent.VK_RIGHT);
				}
			};
			execTimer.schedule(responsePhase, 200);
		}
	}

	/**
	 * Processing of EXECUTION_FINISH phase to define the robot behavior after the end of a session.
	 * Restart the module as many times as defined in the parameter numSessions.
	 */
	private void processExecutionFinishPhase() {
		sessionCounter++;
		if (sessionCounter < numSessions) {
			TimerTask startModuleTask = new TimerTask() {
				public void run() {
					robot.keyPress(KeyEvent.VK_ENTER);
					robot.keyRelease(KeyEvent.VK_ENTER);
				}
			};
			execTimer.schedule(startModuleTask, 200);
		} else {
			TimerTask startModuleTask = new TimerTask() {
				public void run() {
					exportData();
					cleanup();
				}
			};
			execTimer.schedule(startModuleTask, 200);
		}
	}
}
